# Copyright © 2023 Neven Sajko

module ApproximateInfinityNorm

import
  Polynomials,
  ..NumericalErrorTypes,
  ..NumericalErrors,
  ..LinearInterpolation,
  ..SortedBoundedLengthVectors,
  ..GoldenSectionSearch

const Order = Base.Order
const SortedBoundedLengthVector =
  SortedBoundedLengthVectors.SortedBoundedLengthVector
const GSS = GoldenSectionSearch
const NE = NumericalErrorTypes
const numerical_error = NumericalErrors.numerical_error

gss_maximum(fun::Fun, itv::NTuple{2,F}) where {Fun, F<:AbstractFloat} =
  GSS.search(fun, itv, GSS.GSS{GSS.Maximum}())

struct Node{F <: Real}
  # Interval bounds
  interval::NTuple{2, F}

  # Max error among the errors of the upper and lower interval bounds
  error::F
end

Base.isless(l::Node{F}, r::Node{F}) where {F} = Base.isless(l.error, r.error)

const MaxOrder = Base.Order.Reverse
const MaxSortedVec = SortedBoundedLengthVector{Node{F}, typeof(MaxOrder)} where {F}

function interval_max_err_rough!(
  error_type::NE.NumericalErrorType,
  fun::Fun,
  pol::Polynomials.AbstractPolynomial{F},
  max_errs::MaxSortedVec{F},
  interval::NTuple{2, F},
  density1::Int,
  density2::Int,
) where {Fun<:Any, F<:Real}
  d1 = 1 << density1
  d2 = 1 << density2

  lerpi = let m = d1*d2, itv = interval
    i -> LinearInterpolation.lerpi(i, m, itv...)
  end

  a = first(interval)
  b = lerpi(d2)

  i = 0

  while true
    error = zero(F)

    for j ∈ 1:(d2 - 1)
      p = lerpi(i*d2 + j)
      e = numerical_error(error_type, fun(p), pol(p))::F
      (error < e) && (error = e)
    end

    push!(max_errs, Node((a, b), error))
    i += 1
    (d1 < i + 1) && break
    a = b
    b = lerpi((i + 1)*d2)
  end

  nothing
end

is_tiny(::NE.AbsoluteError, ::Real, ::Real) = false

is_tiny(::NE.RelativeError, accurate::Real, maximal_tiny::Real) =
  abs(accurate) < maximal_tiny

function numerical_error_with_status(
  error_type::NE.NumericalErrorType,
  fun::Fun,
  pol::Polynomials.AbstractPolynomial{F},
  position::F,
  maximal_tiny::Real,
) where {Fun<:Any, F<:Real}
  accurate = fun(position)
  approximate = pol(position)
  (
    is_tiny = is_tiny(error_type, accurate, maximal_tiny),
    error = numerical_error(error_type, accurate, approximate)::F,
  )
end

function interval_max_err(
  error_type::NE.NumericalErrorType, fun::Fun, pol::Polynomials.AbstractPolynomial{F},
  interval::NTuple{2, F}, maximal_tiny::Real = 1e-5,
  state_len::Int = 4, density1::Int = state_len + 5, density2::Int = 3,
) where {Fun<:Any, F<:Real}
  rough = MaxSortedVec{F}(1 << state_len, MaxOrder)
  interval_max_err_rough!(error_type, fun, pol, rough, interval, density1, density2)
  max_err = -one(F)
  max_err_nontiny = -one(F)
  position = zero(F)

  error_fun = let f = fun, p = pol
    x -> numerical_error(error_type, f(x), p(x))::F
  end

  while !isempty(rough)
    intrvl = pop!(rough).interval
    itv = gss_maximum(error_fun, intrvl)
    x = first(itv)
    i = 0

    while true
      err_sta = numerical_error_with_status(error_type, fun, pol, x, maximal_tiny)
      e = err_sta.error
      (max_err < e) && (max_err = e)
      if (max_err_nontiny < e) & !err_sta.is_tiny
        max_err_nontiny = e
        position = x
      end
      max_i = 50
      ((max_i < i) || (x == last(itv))) && break
      if i < max_i
        x = nextfloat(x)
      else
        # Too many iterations could mean `itv` is centered around zero.
        x = zero(x)
      end
      i += 1
    end
  end

  (max_err = max_err, max_err_nontiny = max_err_nontiny, position = position)
end

end
