# Copyright © 2022 Neven Sajko. All rights reserved.

module CompressedPolynomials

import Polynomials, MutableArithmetics

const Ps = Polynomials
const MA = MutableArithmetics

# The polynomial is equal to var_x^l * main_factor, where the latter is
# a polynomial in var_y.
struct CompressedPolynomial{
  Coef <: Real,
  var_x,
  var_y,
  P <: Ps.AbstractPolynomial{Coef, var_y},
} <: Ps.AbstractPolynomial{Coef, var_x}

  # Multiplicity of the zero root
  l::Int

  # Change of variable: var_y = var_x^k
  k::Int

  # main factor
  main_factor::P

  function CompressedPolynomial{F, x, y, P}(
    l::Integer,
    k::Integer,
    p::P,
  ) where {x, y, F <: Real, P <: Ps.AbstractPolynomial{F, y}}
    (x isa Symbol) || error("first var is not a Symbol")
    (y isa Symbol) || error("second var is not a Symbol")
    (x == y) && error("variables must be different")
    new{F, x, y, P}(l, k, p)
  end
end

CompressedPolynomial{F, x, y}(
  l::Integer,
  k::Integer,
  p::P,
) where {x, y, F <: Real, P <: Ps.AbstractPolynomial{F, y}} =
  CompressedPolynomial{F, x, y, P}(l, k, p)

CompressedPolynomial{F, x}(
  l::Integer,
  k::Integer,
  p::P,
) where {x, y, F <: Real, P <: Ps.AbstractPolynomial{F, y}} =
  CompressedPolynomial{F, x, y}(l, k, p)

# Copy constructor
CompressedPolynomial{F, x, y, P}(
  p::CompressedPolynomial{F, x, y, P},
) where {x, y, F <: Real, P <: Ps.AbstractPolynomial{F, y}} =
  CompressedPolynomial{F, x}(p.l, p.k, P(p.main_factor))

# TODO: get rid of unnecessary `Val` if the interface gets fixed upstream
Ps.copy_with_eltype(
  ::Val{New},
  p::CompressedPolynomial{Old, x},
) where {New <: Real, Old <: Real, x} =
  CompressedPolynomial{New, x}(p.l, p.k, Ps.copy_with_eltype(Val{New}(), p.main_factor))

function Base.show(
  io::IO,
  p::CP,
) where {var_x, var_y, CP <: CompressedPolynomial{<:Any, var_x, var_y}}
  print(
    io,
    CP, "(change of variable: ", var_y, " = ", var_x, "^", p.k, ", function: ", var_x,
    "^", p.l, " * ",
  )
  show(io, p.main_factor)
  print(io, ")")
end

Base.show(io::IO, ::MIME"text/plain", p::CompressedPolynomial) =
  show(io, p)

Ps.degree(p::CompressedPolynomial) = p.l + p.k * Ps.degree(p.main_factor)

Base.length(p::CompressedPolynomial) = length(p.main_factor)

Base.zero(::Type{CP}) where {
  x, y, P <: Ps.AbstractPolynomial, CP <: CompressedPolynomial{<:Any, x, y, P},
} =
  CP(false, gcd(false), zero(P))

Base.one(::Type{CP}) where {
  x, y, P <: Ps.AbstractPolynomial, CP <: CompressedPolynomial{<:Any, x, y, P},
} =
  CP(false, gcd(false), one(P))

function Base.evalpoly(x::Real, p::CompressedPolynomial)
  y = x^(p.k)
  x^(p.l) * evalpoly(y, p.main_factor)
end

const DenseCP = CompressedPolynomial{F, x, y, Ps.Polynomial{F, y}} where {F, x, y}

function Base.evalpoly(x::F, p::DenseCP{F}) where {F <: Real}
  y = x^(p.k)
  a = x^(p.l)
  v = zero(F)
  v = MA.operate_to!!(v, evalpoly, y, Ps.coeffs(p.main_factor))
  v = MA.operate!!(*, v, a)
end

(p::CompressedPolynomial)(x::Real) = Base.evalpoly(x, p)

function reconstruct_coefficients_from_monomial_indices(
  coefs::AbstractVector{F},
  monomials::AbstractVector{<:Integer},
) where {F <: Any}

  ret = F[]
  j = firstindex(coefs)

  for i ∈ zero(eltype(monomials)):last(monomials)
    c = zero(F)
    if i ∈ monomials
      c = coefs[j]
      j += true
    end
    push!(ret, c)
  end

  ret
end

function CompressedPolynomial{F, var_x, var_y, P}(
  coefs::AbstractVector{F},
  monomials::AbstractVector{<:Integer},
) where {
  F <: Real,
  var_x,
  var_y,
  P <: Ps.Polynomial{F, var_y},
}
  ==(map(length, (coefs, monomials))...) || error("length mismatch")
  issorted(monomials) || error("monomials should be sorted")
  allunique(monomials) || error("monomials can't have duplicates")

  isempty(monomials) && (
    # zero polynomial
    return zero(CompressedPolynomial{F, var_x, var_y, P})
  )

  l = first(monomials)

  subtracter = let l = l
    x -> x - l
  end

  monomials_tmp = map(subtracter, monomials)

  k = gcd(monomials_tmp)

  iszero(k) && (
    # There is just a single monomial
    return CompressedPolynomial{F, var_x, var_y, P}(l, k, P(coefs))
  )

  divider = let k = k
    x -> x ÷ k
  end

  map!(divider, monomials_tmp, monomials_tmp)

  CompressedPolynomial{F, var_x, var_y, P}(
    l,
    k,
    P(reconstruct_coefficients_from_monomial_indices(
      coefs,
      monomials_tmp,
    )),
  )
end

function nonzero_monomials(coefs::AbstractVector, ::Type{T}) where {T <: Integer}
  ret = T[]

  for i ∈ map(T, 0:(length(coefs) - 1))
    iszero(coefs[begin + i]) || push!(ret, i)
  end

  ret
end

adder(n::Number) =
  let n = n
    x -> x + n
  end

function polynomial_to_coefs_and_monomials(
  p::Ps.Polynomial,
  ::Type{T},
) where {T <: Integer}
  co = Ps.coeffs(p)

  adder = let n = T(firstindex(co))
    (x::T) -> x + n
  end

  mo = nonzero_monomials(co, T)

  (co[map(adder, mo)], mo)
end

CompressedPolynomial{F, var_x, var_y, Ps.Polynomial{F, var_y}}(
  p::Ps.Polynomial{F, var_x},
) where {F <: Real, var_x, var_y} =
  CompressedPolynomial{F, var_x, var_y, Ps.Polynomial{F, var_y}}(
    polynomial_to_coefs_and_monomials(p, UInt8)...,
  )

end
