# Copyright © 2023 Neven Sajko

# Derivative-free optimization for a unidimensional unimodal function.

module GoldenSectionSearch

import ..NumericalErrorTypes, ..NumericalErrors

abstract type Objective end
struct Minimum <: Objective end
struct Maximum <: Objective end

(::Minimum)() = <
(::Maximum)() = >

struct GSS{O<:Objective}
  abs::Int
end

GSS{O}() where {O<:Objective} = GSS{O}(-5000)

absolute_tolerance(::Type{F}, gss::GSS) where {F<:AbstractFloat} = ldexp(one(F), gss.abs)

strict_order(::GSS{O}) where {O} = O()()

invφ( ::Type{T}) where {T<:AbstractFloat} = (sqrt(T(0x5)) - 1)/2  # 1/φ
invφ2(::Type{T}) where {T<:AbstractFloat} = (3 - sqrt(T(0x5)))/2  # 1/φ^2

invφ(::Type{Float32}) = 0.618034f0
invφ(::Type{Float64}) = 0.6180339887498949

invφ2(::Type{Float32}) = 0.38196602f0
invφ2(::Type{Float64}) = 0.38196601125010515

function search(
  fun::Fun, ab::NTuple{2, F}, gss::GSS,
) where {Fun<:Function, F<:AbstractFloat}
  # Roughly based on:
  # https://en.wikipedia.org/w/index.php?title=Golden-section_search&oldid=1135077979#Iterative_algorithm

  iφ = invφ(F)
  iφ2 = invφ2(F)
  ord = strict_order(gss)
  (a, b) = minmax(ab...)
  h = b - a

  c = fma(iφ2,  h, a)
  d = fma(iφ2, -h, b)

  (a < c < d < b) || (return (a, b))

  (yc, yd) = map(fun, (c, d))
  abs_tol = absolute_tolerance(F, gss)

  while (abs_tol < h) && (a < c < d < b)
    h *= iφ
    o = ord(yc, yd)
    n = fma(iφ2, (o ? (h, a) : (-h, b))...)
    (
      ( o & (a < n < c)) |
      (!o & (d < n < b))
    ) || break

    if o
      b = d
      d = c
      yd = yc
      c = n
      yc = fun(c)
    else
      a = c
      c = d
      yc = yd
      d = n
      yd = fun(d)
    end
  end

  ifelse(ord(yc, yd), (a, d), (c, b))
end

function single(f::Fun, ab::NTuple{2, F}, gss::GSS{Maximum}) where {Fun,F<:AbstractFloat}
  (l, r) = search(f, ab, gss)
  if l < false < r
    let z = zero(l)
      (z, f(z))
    end
  else
    let fl = f(l), fr = f(r)
      (fl < fr) ? (r, fr) : (l, fl)
    end
  end::Tuple{F,Any}
end

end
