# Copyright © 2023 Neven Sajko

module Minimax

import
  MathOptInterface,
  Polynomials,
  ..NumericalErrorTypes,
  ..NumericalErrors,
  ..PolynomialPassingThroughIntervals,
  ..LinearInterpolation,
  ..PolynomialApproximationErrorLocalMaxima,
  ..IntervalHelpers,
  ..CompressedPolynomials,
  ..TinyFilter

const MOI = MathOptInterface
const NE = NumericalErrorTypes
const PPTI = PolynomialPassingThroughIntervals
const ItvH = IntervalHelpers
const PAELM = PolynomialApproximationErrorLocalMaxima
const numerical_error = NumericalErrors.numerical_error
const local_maxima = PAELM.local_maxima
const tiny_filter = TinyFilter.tiny_filter
const evenly_spaced_grid = LinearInterpolation.evenly_spaced_grid
const polynomial_pti! = PPTI.polynomial_passing_through_intervals!
const ComPol = CompressedPolynomials.CompressedPolynomial

const CP = ComPol{F, :x, :y, Polynomials.Polynomial{F, :y}} where {F}

struct InsufficientLPSamplePointsException <: Exception
  lp_sample_point_count::Int
  monomial_count::Int
end

struct LPSamplePointNotFiniteException <: Exception end

function mmx_single(
  make_lp::MakeLP,
  f::Fun,
  intervals::AbstractVector{NTuple{2,F}},
  exact_points,
  grids::AbstractVector{<:AbstractVector{F}},
  monomials::AbstractVector{<:Integer},
  error_type::NE.NumericalErrorType,
  ::LPOption,
  max_error_upper_bound::F,
  mmx_error_roots_granularity::Int,
) where {MakeLP, Fun, F<:AbstractFloat, LPOption<:PPTI.FeasibleOption}
  let l_lp = mapreduce(length, +, grids, init = 0), l_mo = length(monomials)
    (l_lp < l_mo) && throw(InsufficientLPSamplePointsException(l_lp, l_mo))
  end
  foreach(
    (g -> foreach((x -> (isfinite(x) || throw(LPSamplePointNotFiniteException()))), g)),
    grids,
  )

  lp_grid = reduce(vcat, grids, init = F[])
  func_vals = map(f, lp_grid)
  lp_intervals = ItvH.intervals_given_max_err(func_vals, error_type, max_error_upper_bound)
  append!(lp_grid, exact_points)
  let
    exact_vals = map(f, exact_points)
    append!(func_vals, exact_vals)
    for v ∈ exact_vals
      push!(lp_intervals, (v, v))
    end
  end
  ppti_minimax_option =
    PPTI.MinimaxOption{LPOption}(func_vals, error_type, max_error_upper_bound)
  coefs = polynomial_pti!(
    make_lp(), monomials, lp_grid, lp_intervals, F, ppti_minimax_option,
  )

  polynomial = if !isempty(coefs)
    CP{F}(coefs, monomials)
  else
    CP{F}(coefs, true:false)
  end

  maxima_of_error = let f = f, pol = polynomial, gran = mmx_error_roots_granularity
    (itv::NTuple{2,F}) -> local_maxima(error_type, f, pol, itv, gran)
  end

  error_maxima = if !isempty(coefs)
    map(maxima_of_error, intervals)
  else
    Vector{NTuple{2,F}}[]
  end::Vector{Vector{NTuple{2,F}}}

  (; coefs, polynomial, error_maxima)
end

to_function(fun::Fun, ::Type{Ret}) where {Fun, Ret} = let f = fun
  x -> f(x)::Ret
end

deep_reduce(op::Op, c::AbstractArray                 ) where {Op} = mapreduce(last,            op, c)
deep_reduce(op::Op, c::AbstractArray{<:AbstractArray}) where {Op} = mapreduce(deep_reduce(op), op, c)
deep_reduce(op::Op) where {Op} = let op = op
  (e::AbstractArray) -> deep_reduce(op, e)
end

interval_length(itv::NTuple{2,F}) where {F<:Real} = last(itv) - first(itv)

min_interval_length(itvs::Vector{NTuple{2,F}}) where {F<:Real} =
  mapreduce(interval_length, min, itvs)

"""
`min_interval_sample_point_count` is the number of sample points that the
interval of smallest length should have. Other intervals get proportionally
more sample points.
"""
function init_lp_grids(
  intervals::Vector{NTuple{2,F}}, min_interval_sample_point_count::Int,
) where {F<:AbstractFloat}
  itv_len_fac = let l = interval_length, min_len = min_interval_length(intervals)
    (itv::NTuple{2,F}) -> l(itv) / min_len
  end

  gridder_impl = let c = min_interval_sample_point_count
    (fac::Real, itv::NTuple{2,F}) -> evenly_spaced_grid(ceil(Int, fac*c), itv...)
  end

  gridder = let g = gridder_impl, ilf = itv_len_fac
    (itv::NTuple{2,F}) -> g(ilf(itv), itv)
  end

  Vector{F}[gridder(itv) for itv ∈ intervals]
end

function minimax_polynomial_raw_intervals(
  make_lp::MakeLP, fun::Fun, intervals::Vector{NTuple{2,F}},
  monomials::AbstractVector{<:Integer}, error_type::NE.NumericalErrorType,
  lp_option::LPOption, mmx_error_roots_granularity::Int, exact_points,
  debug::Debug,
  lp_grid_initializer::GridIni = init_lp_grids,
) where {MakeLP, Fun, F<:AbstractFloat, LPOption<:PPTI.FeasibleOption, Debug, GridIni}
  find_mmx = let ma = make_lp, f = fun, i = intervals, ep = exact_points
    mo = monomials
    lpo = lp_option
    g = mmx_error_roots_granularity
    (grids::AbstractVector{<:AbstractVector{F}}, meub::F) ->
      mmx_single(ma, f, i, ep, grids, mo, error_type, lpo, meub, g)
  end

  result = let grids = lp_grid_initializer(intervals, length(monomials) + 3)
    mmx = find_mmx(grids, typemax(F))
    max_error_upper_bound = deep_reduce(max, mmx.error_maxima)
    (; mmx, max_error_upper_bound)
  end

  grids_from_errmax = errmax -> map((i -> map(first, i)), errmax)
  deb = let i = intervals, f = fun, d = debug
    (iter, mmx) -> d(iter, i, f, mmx)
  end
  iter = 0
  deb(iter, result.mmx)

  isempty(result.mmx.coefs) || while true
    grids = grids_from_errmax(result.mmx.error_maxima)

    # leave some wiggle room for numerical error in creating the intervals
    max_max = nextfloat(result.max_error_upper_bound, 10)

    mmx = find_mmx(grids, max_max)
    isempty(mmx.coefs) && break
    iter += true
    deb(iter, mmx)
    max_error_upper_bound = deep_reduce(max, mmx.error_maxima)

    # TODO: maybe compare against the current lower bound instead of against
    #       the previous upper bound?
    if max_error_upper_bound < result.max_error_upper_bound
      result = (; mmx, max_error_upper_bound)
    else
      break
    end
  end

  mmx = result.mmx
  max_error = (
    deep_reduce(min, mmx.error_maxima),
    result.max_error_upper_bound,
  )::NTuple{2, F}
  (; intervals, mmx, max_error)
end

"""
    minimax_options(options...)

Options for [`minimax_polynomial`](@ref).

  * The type of the error that we're minimizing, absolute or relative

    * Optional argument

  * Options for the LP formulation

    * Optional argument

  * The granularity used during the root-finding bracketing when filtering tiny-valued
    subintervals

    * Keyword argument

  * The granularity used during the root-finding bracketing when finding maxima of the
    approximation error

    * Keyword argument

  * Maximal value considered to be too "tiny" for LP

    * Keyword argument

  * A callable object that may be used for debugging

    * Keyword argument
"""
minimax_options(
  error_type::NE.NumericalErrorType = NE.RelativeError(),
  lp_option::LPOption = PPTI.ReplacedVariablesOption{:determinants}();
  tiny_filter_roots_granularity::Int = 15,
  mmx_error_roots_granularity::Int = 11,
  maximal_tiny::Real = 1e-5,
  exact_points = (),
  debug::Debug = Returns(nothing),
) where {LPOption<:PPTI.FeasibleOption, Debug} = (;
  error_type, lp_option, tiny_filter_roots_granularity, mmx_error_roots_granularity,
  maximal_tiny, exact_points, debug,
)

"""
    minimax_polynomial(make_lp, fun, intervals, monomials, options =  minimax_options())

Finds a minimax polynomial.

Arguments:

  * A function that returns a new MOI.AbstractOptimizer

    * A new MathOptInterface LP optimizer each time it's called
    * Only one will be used at a time
    * Tested with Tulip

  * A function to be approximated

    * Only univariate real functions are supported

  * A vector of `NTuple{2, F}` elements, each of which represents an interval
    over which the given function is to be approximated. The error is minimized over all
    intervals simultaneously.

    * `F` subtypes AbstractFloat
    * The first value in an interval is the lower bound of the interval, and the second
      value is the upper bound
    * None of the two bounds of an interval may be a root of the given
      function when the error type is relative error. If this is a problem,
      the user should try slightly shrinking or expanding the interval over
      the root.

  * The indices denoting the monomials that the sought-after polynomial is
    supposed to have

    * For example, use `0:3` to look for any polynomial of degree three, or
      `0:2:8` for any even polynomial of degree eight
    * Note: the user should use even polynomials for approximating even
      functions and odd polynomials for approximating odd functions

  * Optional options, see [`minimax_options`](@ref).

Returns a named tuple with these fields:

  * `intervals`: just like the intervals given as input, but filtered from tiny-valued
    subintervals

  * `mmx`: the polynomial

  * `max_error::NTuple{2,AbstractFloat}`: the lower and upper bounds on the maximum
    approximation error

See the package help entry for a usage example:
```
import FindMinimaxPolynomial
?FindMinimaxPolynomial
```
"""
function minimax_polynomial(
  make_lp::MakeLP,
  fun::Fun,
  intervals::Vector{NTuple{2,F}},
  monomials::AbstractVector{<:Integer},
  options::Options = minimax_options(),
) where {MakeLP, Fun, F<:AbstractFloat, Options}
  mt = options.maximal_tiny
  tfrg = options.tiny_filter_roots_granularity
  signbit(mt) && error("negative magnitude")
  (tfrg < 1) && error("granularity must be positive")
  f = to_function(fun, F)
  refined_itvs = tiny_filter(f, intervals, mt, tfrg)
  minimax_polynomial_raw_intervals(
    make_lp, fun, refined_itvs, monomials, options.error_type, options.lp_option,
    options.mmx_error_roots_granularity, options.exact_points, options.debug,
  )
end

end
