# Copyright © 2023 Neven Sajko

module PolynomialApproximationErrorLocalMaxima

import
  ..NumericalErrorTypes,
  ..NumericalErrors,
  ..FindRoots,
  ..GoldenSectionSearch,
  ..CollectionWithReserved

const NE = NumericalErrorTypes
const numerical_error = NumericalErrors.numerical_error
const coll_reserved = CollectionWithReserved.collection_with_reserved

function unimodal_maximum(fun::Fun, itv::NTuple{2, F}) where {Fun, F<:AbstractFloat}
  GSS = GoldenSectionSearch
  gss = GSS.GSS{GSS.Maximum}()
  GSS.single(fun, itv, gss)
end

function local_maxima(
  et::NE.NumericalErrorType,
  f::Acc,
  pol::App,
  interval::NTuple{2,F},
  roots_granularity::Int,
) where {Acc, App, F}
  # Between two consecutive roots of the polynomial approximation error, the
  # error seems to be unimodal, so GSS can be used to find the maximum.

  # Use signed error so we could find its roots by bracketing.
  signed_err = let et = signed(et), f = f, pol = pol
    (x::F) -> numerical_error(et, f(x), pol(x))::F
  end

  u_max = let err = abs ∘ signed_err
    (itv::NTuple{2,F}) -> unimodal_maximum(err, itv)
  end

  rts = FindRoots.find_roots(signed_err, interval, roots_granularity)

  rt_cnt = length(rts)

  ret = coll_reserved(Vector{NTuple{2,F}}, rt_cnt + 1)

  if !isempty(rts)
    push!(ret, u_max((first(interval), first(rts))))

    for i ∈ 1:(rt_cnt - 1)
      push!(ret, u_max((rts[begin + i - 1], rts[begin + i])))
    end

    push!(ret, u_max((last(rts), last(interval))))
  else
    push!(ret, u_max(interval))
  end

  ret
end

function interval_max_err(
  et::NE.NumericalErrorType,
  f::Acc,
  pol::App,
  interval::NTuple{2,F},
  roots_granularity::Int,
) where {Acc, App, F}
  my_max = c -> last(sort!(c, by = last))
  (x, y) = my_max(local_maxima(et, f, pol, interval, roots_granularity))
  (max_err = y, position = x)
end

end
