# Copyright © 2022 Neven Sajko

module PolynomialPassingThroughIntervals

import MathOptInterface

const MOI = MathOptInterface

using
  ..NumericalErrorTypes,
  ..LinCombs,
  ..SymbolicGaussianEliminationVandermonde,
  ..OverdeterminedSystemConsistencyConstraints,
  ..VandermondeInverse

two(::Type{T}) where {T <: Number} = T(0x2)

half(::Type{T}) where {T <: Number} = inv(two(T))

# Interval
const Itv = NTuple{2, F} where {F <: Real}

has_nan(i::Itv{<:Real}) = any(isnan, i)
is_empty(i::Itv{<:Real}) = >(i...)

function check_validity_of_intervals(is::AbstractVector{<:Itv{<:Real}})
  any(has_nan, is) && error("interval has NaN endpoint")
  any(is_empty, is) && error("empty interval detected")
end

linear_function(
  vars::Vector{MOI.VariableIndex},
  ::Type{F},
  lc::LinComb{<:Real}) where {F <: AbstractFloat} =
  MOI.ScalarAffineFunction(
    map(
      let vars = vars
        term -> MOI.ScalarAffineTerm(F(last(term)), vars[first(term).h])
      end,
      lc),
    zero(F))

const NonemptyNTuple = Tuple{T, Vararg{T, n}} where {n, T}

linear_function(
  vars::NonemptyNTuple{n, MOI.VariableIndex},
  coeffs::NonemptyNTuple{n, F}) where {n, F <: AbstractFloat} =
  MOI.ScalarAffineFunction(
    collect(map(
      ((c, var) -> MOI.ScalarAffineTerm(F(c), var)),
      coeffs,
      vars)),
    zero(F))

struct Interpolater{T <: Number}
  i::Matrix{T}

  function Interpolater(m::Matrix{T}) where {T <: Number}
    ==(size(m)...) || error("the matrix must be square, given size: ", size(m))
    new{T}(m)
  end
end

(i::Interpolater)(vals) = i.i * vals

Interpolater{T}(m::Vector{<:LinComb{<:Any}}) where {T <: Any} =
  Interpolater(to_dense(T, m))

abstract type AbstractArgHelper end

monomials_are_full(m::AbstractVector{T}) where {T <: Integer} =
  ==(m, map(T, 0:(length(m) - 1)))

abstract type AbstractOption end

abstract type FeasibleOption <: AbstractOption end

struct ReplacedVariablesOption{sym} <: FeasibleOption
  function ReplacedVariablesOption{sym}() where {sym}
    in(sym, (:gaussian_elimination, :determinants)) || error("unrecognized option")
    new{sym}()
  end
end

struct MinimaxOption{
  O <: FeasibleOption,
  F <: Real,
  E <: NumericalErrorType,
} <: AbstractOption

  accurate_values::Vector{F}

  max_max_error::F
end

MinimaxOption{O}(
  accurate_values::Vector{F},
  et::E = RelativeError(),
  max_max_error::F = typemax(F),
) where {O <: FeasibleOption, F <: Real, E <: NumericalErrorType} =
  MinimaxOption{O, F, E}(accurate_values, max_max_error)

error_type(::MinimaxOption{<:Any, <:Any, E}) where {E <: NumericalErrorType} =
  E()

struct ArgHelperReplacedVariables{
  Interp <: AbstractFloat,
  Constr <: Real} <: AbstractArgHelper

  intp::Interpolater{Interp}
  constraints::Vector{LinComb{Constr}}

  function ArgHelperReplacedVariables{I}(
    monomials::AbstractVector{<:Integer},
    poldom::AbstractVector{F},
    ::ReplacedVariablesOption{:gaussian_elimination}) where
  {I <: AbstractFloat, F <: Real}
    local t = vandermonde_important_parts(monomials, poldom)
    new{I, F}(Interpolater{I}(t.solution), t.constraints)
  end

  function ArgHelperReplacedVariables{I}(
    monomials::AbstractVector{<:Integer},
    poldom::AbstractVector{F},
    ::ReplacedVariablesOption{:determinants}) where
  {I <: AbstractFloat, F <: Real}
    local i = zeros(F, 0, 0)
    local c = LinComb{F}[]
    if monomials_are_full(monomials)
      local degree = length(monomials) - 1
      i = vandermonde_inv(@view poldom[begin:(begin - 1 + length(monomials))])
      c = overdetermined_consistency_constraints_vandermonde(degree, poldom)
    else
      i = inv(
        F[(poldom[begin - 1 + row])^e
          for row in 1:length(monomials), e in monomials])
      c = overdetermined_consistency_constraints_general(monomials, poldom)
    end
    new{I, F}(Interpolater(map(I, i)), c)
  end
end

struct ArgHelperMinimax{
  E <: NumericalErrorType,
  F <: Real,
  AH <: AbstractArgHelper,
} <: AbstractArgHelper

  error_type::E

  accurate_values::Vector{F}

  max_max_error::F

  ah::AH
end

error_type(::ArgHelperMinimax{E}) where {E <: NumericalErrorType} =
  E()

arg_helper(
  ::Type{I},
  monomials::AbstractVector{<:Integer},
  poldom::AbstractVector{<:Real},
  o::ReplacedVariablesOption) where {I <: Any} =
  ArgHelperReplacedVariables{I}(monomials, poldom, o)

arg_helper(
  ::Type{I},
  monomials::AbstractVector{<:Integer},
  poldom::AbstractVector{<:Real},
  o::MinimaxOption{RV},
) where {I <: Any, RV <: ReplacedVariablesOption} =
  ArgHelperMinimax(
    error_type(o),
    o.accurate_values,
    o.max_max_error,
    arg_helper(I, monomials, poldom, RV()))

function lp_configure!(
  lp::MOI.AbstractOptimizer,
  ah::ArgHelperReplacedVariables{<:AbstractFloat, <:Real},
  target_intervals::Vector{Itv{F}}) where {F <: AbstractFloat}

  MOI.set(lp, MOI.ObjectiveSense(), MOI.FEASIBILITY_SENSE)

  lp_vars = Vector{MOI.VariableIndex}()

  for itv ∈ target_intervals
    var = if ==(itv...)
      let
        (v, _) = MOI.add_constrained_variable(lp, MOI.EqualTo(first(itv)))
        v
      end
    else
      let
        (v, _) = MOI.add_constrained_variable(lp, MOI.Interval(itv...))
        v
      end
    end
    push!(lp_vars, var)
  end

  for constr in ah.constraints
    MOI.add_constraint(
      lp,
      linear_function(lp_vars, F, constr),
      MOI.EqualTo(zero(F)))
  end

  lp_vars
end

# `abs(acc - app) ≤ err` is equivalent to `-err ≤ acc - app ≤ err`, which is
# equivalent to the conjuction of
#
#   `app - err ≤ acc`
#
# and
#
#   `-app - err ≤ -acc`
error_interval_inclusion(
  ::AbsoluteError,
  accurate_value::F,
  approximate_value::V,
  error::V,
) where {F <: Real, V <: MOI.VariableIndex} =
  let vs = (approximate_value, error), acc = accurate_value
    (
      # `app - err ≤ acc`
      (
        linear_function(vs, (one(F), F(-one(Int8)))),
        MOI.LessThan{F}(acc),
      ),

      # `-app - err ≤ -acc`
      (
        linear_function(vs, (F(-one(Int8)), F(-one(Int8)))),
        MOI.LessThan{F}(-acc),
      ),
    )
  end

# `abs((acc - app)/acc) ≤ err` is equivalent to `-err ≤ (acc - app)/acc ≤ err`,
# which is equivalent to the conjuction of
#
#   `(app - acc)/acc ≤ err`
#
# and
#
#   `(acc - app)/acc ≤ err`,
#
# which is equivalent to the conjuction of
#
#   `(app - acc - err)/acc ≤ 0`
#
# and
#
#   `(acc - app - err)/acc ≤ 0`,
#
# which is equivalent to the conjuction of
#
#   `sgn(acc)*app - abs(acc)*err ≤ abs(acc)`
#
# and
#
#   `-sgn(acc)*app - abs(acc)*err ≤ -abs(acc)`
error_interval_inclusion(
  ::RelativeError,
  accurate_value::F,
  approximate_value::V,
  error::V,
) where {F <: Real, V <: MOI.VariableIndex} =
  let vs = (approximate_value, error), a = abs(accurate_value), s = sign(accurate_value)
    (
      # `sgn(acc)*app - abs(acc)*err ≤ abs(acc)`
      (
        linear_function(vs, (s, -a)),
        MOI.LessThan{F}(a),
      ),

      # `-sgn(acc)*app - abs(acc)*err ≤ -abs(acc)`
      (
        linear_function(vs, (-s, -a)),
        MOI.LessThan{F}(-a),
      ),
    )
  end

function lp_configure!(
  lp::MOI.AbstractOptimizer,
  ah::ArgHelperMinimax{E, F, <:ArgHelperReplacedVariables},
  target_intervals::Vector{Itv{F}},
) where {E <: NumericalErrorType, F <: AbstractFloat}

  local lp_vars = lp_configure!(lp, ah.ah, target_intervals)

  MOI.set(lp, MOI.ObjectiveSense(), MOI.MIN_SENSE)

  local max_error = MOI.add_variable(lp)

  MOI.add_constraint(
    lp,
    max_error,
    MOI.Interval(zero(F), ah.max_max_error))

  MOI.set(
    lp,
    MOI.ObjectiveFunction{MOI.ScalarAffineFunction{F}}(),
    MOI.ScalarAffineFunction([MOI.ScalarAffineTerm(one(F), max_error)], zero(F)))

  for (acc, approx_val) in zip(ah.accurate_values, lp_vars)
    local eic = error_interval_inclusion(
      error_type(ah), acc, approx_val, max_error)
    for constraint in eic
      MOI.add_constraint(lp, constraint...)
    end
  end

  lp_vars
end

moi_termination_status(lp::MOI.AbstractOptimizer) = MOI.get(lp, MOI.TerminationStatus())

moi_termination_status_is_optimal(lp::MOI.AbstractOptimizer) =
  moi_termination_status(lp) == MOI.OPTIMAL

moi_vars(lp::MOI.AbstractOptimizer, vars::Vector{MOI.VariableIndex}) =
  MOI.get(lp, MOI.VariablePrimal(), vars)

function lp_optimize!(
  ::Type{F}, lp::MOI.AbstractOptimizer, vars::Vector{MOI.VariableIndex},
) where {F<:Real}
  throwed = try
    MOI.optimize!(lp)
    false
  catch
    true
  end::Bool

  if !throwed && moi_termination_status_is_optimal(lp)
    moi_vars(lp, vars)
  else
    F[]
  end::Vector{F}
end

interpolater(ah::ArgHelperReplacedVariables) = ah.intp

interpolater(ah::ArgHelperMinimax) = interpolater(ah.ah)

function polynomial_passing_through_intervals!(
  lp::MOI.AbstractOptimizer,
  ah::AbstractArgHelper,
  target_intervals::Vector{Itv{S}}) where {S <: AbstractFloat}

  check_validity_of_intervals(target_intervals)

  local lp_vars = lp_configure!(lp, ah, target_intervals)
  local lp_result = lp_optimize!(
    S,
    lp,
    lp_vars[begin:(begin - 1 + size(interpolater(ah).i, 2))])

  isempty(lp_result) && (return S[])

  interpolater(ah)(lp_result)
end

"""
Tries to find a univariate polynomial of chosen degree that passes
through the given intervals. The returned polynomial is represented
with its coefficients.

The degree is not given explicitly, rather the user provides a
collection of monomial indices. This requirement to explicitly select
chosen monomial positions allows the user to "zero" some monomials.
For example, providing [0, 2, 3] as the monomials indicates a wish
for a polynomial of this form: a_0 + a_2*x^2 + a_3*x^3

Note: when the monomial `0` is not included in monomials, it is
preferable not to include zero in poldom. The zeroth monomial is the
constant term of the polynomial, and any polynomial without a constant
term maps zero to zero anyway.

Note: including only odd monomials guarantees an odd polynomial (such
as might be desirable for approximating the sine), while including only
even polynomials guarantees an even polynomial (which might be used for
an even function like the cosine).

Returns an empty vector of coefficients when no polynomial is found.

polynomial_domain must be a finite sorted collection of real
numbers. Its element type must be convertible to Rational{BigInt}.

Multiple linear programming formulations may be used to seek for the
solution:

 * `ReplacedVariablesOption{:determinants}` - the default option,
   should be the fastest and most numerically accurate. The other
   options are mostly here for comparison. Finds a feasible polynomial.
   The linear optimization looks for feasible points in the given
   intervals for the polynomial curve to pass through. Afterwards the
   sought-after polynomial is interpolated as a linear function of
   the first degree+1 points. This interpolation is done using
   InterpolationType, BigFloat by default. Ensure the precision of
   BigFloat or any other chosen type is set to a suitably high value,
   if necessary.

 * `ReplacedVariablesOption{:gaussian_elimination}` - the same LP
   formulation as above, but the preparation of the formulation may be
   faster for really large polynomials (measured in number of
   monomials/coefficients). Requires much higher numerical precision
   than the `:determinants` variant.

 * `MinimaxOption` - among the polynomials that pass through
   target_intervals over poldom, finds a polynomial that minimizes the
   maximal error with respect to the accurate_values. LP variables are
   the max error and the values of the polynomial. Two constraints are
   added for each value in accurate_values. The maximal maximal error
   can be chosen.

The below options aren't implemented currently:

 * `NaiveOption` - finds a feasible polynomial. LP variables are the
   coefficients of the polynomial. There's a constraint for each
   interval.

 * `AdditionalVariablesOption` - like naive, but with each constraint
   now a variable bounded from both sides.
"""
function polynomial_passing_through_intervals!(
  lp::MOI.AbstractOptimizer,
  monomials::AbstractVector{<:Integer},
  polynomial_domain::AbstractVector{<:Real},
  target_intervals::Vector{Itv{F}},
  ::Type{InterpolationType} = F,
  option::AbstractOption = ReplacedVariablesOption{:determinants}()) where
{F <: AbstractFloat, InterpolationType <: AbstractFloat}

  ==(map(length, (polynomial_domain, target_intervals))...) ||
    error("polynomial_domain and target_intervals lengths must match")

  check_validity_of_intervals(target_intervals)

  polynomial_passing_through_intervals!(
    lp,
    arg_helper(InterpolationType, monomials, polynomial_domain, option),
    target_intervals)
end

end
