# Copyright 2022 Neven Sajko. All right reserved.

module SymbolicGaussianEliminationVandermonde

using LinearAlgebra, ..LinCombs

export vandermonde_important_parts

struct AugmentedMatrix{Coef <: Real}
  mat::Matrix{Coef}

  vec::Vector{LinComb{Coef}}

  AugmentedMatrix(m::Matrix{F}) where {F <: Real} =
    new{F}(m, variables(F, size(m, 1)))
end

function multiply_row!(
  m::AugmentedMatrix{F},
  row::Int,
  c::Real) where {F <: Real}
  # Assuming c is nonzero.

  # Multiply matrix row.
  for i in axes(m.mat, 2)
    m.mat[row, i] *= c
  end

  # Multiply vector element.
  local lin = m.vec[row]
  for i in eachindex(lin)
    lin[i] = first(lin[i]) => (c * last(lin[i]))
  end

  nothing
end

function divide_row!(
  m::AugmentedMatrix{F},
  row::Int,
  c::Real) where {F <: Real}

  iszero(c) && error("division by zero")
  multiply_row!(m, row, inv(c))
end

function normalize!(lin::LinComb{<:Real})
  if 1 < length(lin)
    sort!(lin, by = first)
    local j = lastindex(lin) - 1
    while 1 <= j
      if first(lin[j]) == first(lin[j + 1])
        # These two terms share the variable: merge them.

        local sum = last(lin[j]) + last(lin[j + 1])
        if iszero(sum)
          deleteat!(lin, j:(j + 1))
        else
          deleteat!(lin, j + 1)
          lin[j] = first(lin[j]) => sum
        end

        j -= 1
      end

      j -= 1
    end
  end

  nothing
end

function add_scaled_row!(
  m::AugmentedMatrix{F},
  dstrow::Int,
  srcrow::Int,
  c::Real) where {F <: Real}

  iszero(c) && (return nothing)

  # Add to matrix row.
  for i in axes(m.mat, 2)
    m.mat[dstrow, i] += c * m.mat[srcrow, i]
  end

  # Add to vector element.

  local lin = m.vec[dstrow]

  # Append scaled src to dst.
  for (h, v) in m.vec[srcrow]
    push!(lin, h => (c*v))
  end

  # Merge
  normalize!(lin)

  nothing
end

subtract_scaled_row!(
  m::AugmentedMatrix{F},
  dstrow::Int,
  srcrow::Int,
  c::Real) where {F <: Real} =
   add_scaled_row!(m, dstrow, srcrow, -c)

function nextfl(x::AbstractFloat, ::Val{m}) where {m}
  if 0 <= m
    for i in 1:m
      x = nextfloat(x)
    end
  else
    for i in 1:-m
      x = prevfloat(x)
    end
  end
  x
end

is_almost_zero(x::Rational) = iszero(x)
is_almost_one(x::Rational) = isone(x)

is_almost_zero(x::F) where {F <: AbstractFloat} =
  let small() = nextfl(floatmin(F), Val{10000}())
    abs(x) <= small()
  end

is_almost_one(x::F) where {F <: AbstractFloat} =
  abs(x - 1) <= eps(F)

is_almost_zero(x::AbstractArray) = all(is_almost_zero, x)

function is_almost_one(m::AbstractMatrix)
  ==(size(m)...) || (return false)

  for i in axes(m, 1)
    is_almost_one(m[i, i]) || (return false)
  end

  for i in axes(m, 1)
    for j in axes(m, 2)
      ((i != j) & !is_almost_zero(m[i, j])) && (return false)
    end
  end

  true
end

function is_almost_triu(m::AbstractMatrix)
  for i in axes(m, 1)
    for j in axes(m, 2)
      (<(j, i) & !is_almost_zero(m[i, j])) && (return false)
    end
  end

  true
end

function row_echelon_form!(a::AugmentedMatrix{<:Real})
  local (row_count, col_count) = size(a.mat)

  (col_count + 1 < row_count) || error("unexpected")

  for r in 1:col_count
    # Set leading coefficient to one.
    divide_row!(a, r, a.mat[r, r])

    # Zero the elements below the leading coefficient.
    for m in (r + 1):row_count
      subtract_scaled_row!(a, m, r, a.mat[m, r])
    end
  end

  local s = col_count
  all(is_almost_one, diag(@view a.mat[1:s, 1:s])) ||
    error("internal error top-left id")
  nothing
end

function reduce!(a::AugmentedMatrix{<:Real})
  local (row_count, col_count) = size(a.mat)

  (col_count + 1 < row_count) || error("unexpected")

  # Produce reduced row-echelon form.
  for r in col_count:-1:1
    # Zero the elements above the leading coefficent.
    for m in 1:(r - 1)
      subtract_scaled_row!(a, m, r, a.mat[m, r])
    end
  end

  local s = col_count
  nothing
end

function rref!(a::AugmentedMatrix{<:Real})
  row_echelon_form!(a)
  reduce!(a)
  nothing
end

function important_parts(a::AugmentedMatrix{<:Real})
  local s = size(a.mat, 2)

  local sol = a.vec[1:s]
  local zer = a.vec[(s + 1):end]

  (solution = sol, constraints = zer)
end

function vandermonde(monomials::AbstractVector{<:Integer}, poldom)
  local ret = Matrix{eltype(poldom)}(undef, length(poldom), length(monomials))

  local row = 1
  for x in poldom
    for (i, e) in enumerate(monomials)
      ret[row, i] = x^e
    end

    row += 1
  end

  ret
end

augmented_vandermonde(monomials::AbstractVector{<:Integer}, poldom) =
  AugmentedMatrix(vandermonde(monomials, poldom))

sanity_check(::Any, ::NamedTuple) = nothing

function correct_constraint_format(
  lc::LinComb{F},
  deg::Int,
  var::Int) where {F <: Real}
  (length(lc) == deg + 2) || error("internal error 1")

  for i in 1:(deg + 1)
    (first(lc[i]).h == i) || error("internal error 2")
  end

  (first(last(lc)).h == var) || error("internal error 3")
  is_almost_one(last(last(lc))) || error("internal error 4")

  nothing
end

function sanity_check(::R, r::T) where
{R <: Union{UnitRange, StepRange}, 
 T <: NamedTuple}
  all(lc -> all(isinteger ∘ last, lc), r.constraints) ||
    error("internal error integral linear combination coefficients")

  local deg = length(r.solution) - 1
  for i in 1:length(r.constraints)
    correct_constraint_format(r.constraints[i], deg, deg + 1 + i)
  end

  nothing
end

# eltype(poldom) should be something like Rational{BigInt} or BigFloat.
function vandermonde_important_parts(monomials::AbstractVector{<:Integer}, poldom)
  local m = augmented_vandermonde(monomials, poldom)

  rref!(m)

  local ret = important_parts(m)

  sanity_check(poldom, ret)

  ret
end

module Test

using ..SymbolicGaussianEliminationVandermonde
using ..SymbolicGaussianEliminationVandermonde: LinComb

export some_test_rational

function some_test(max_degree, grids)
  for deg in 1:max_degree
    for grid in grids
      println(deg, " ", grid)

      local t = vandermonde_important_parts(deg, grid)
      local s = t.solution
      local c = t.constraints
      println(first(s), " ", last(s))
      println(first(c), " ", last(c))

      for i in 1:7
        GC.gc(true)
      end

      println()
    end
  end
end

maxint(n) = 2^n - 1

rat(r) = map(Rational{BigInt}, r)

some_test_rational() =
  some_test(28, map(rat, map(n -> 0:maxint(n), 6:16)))

end  # module Test

end
