# Copyright © 2022 Neven Sajko

module VandermondeInverse

using ..SkipVectors

export vandermonde_inv, vandermonde_inv!

# Equivalent to `Int8(-1)^n` and `Int8(iseven(n) ? 1 : -1)`
pow_minus_one(n::Integer) =
  iseven(n)*Int8(2) - one(Int8)

# TODO: MutableArithmetics
subtracter(n::Number) =
  let n = n
    (y::Number) -> n - y
  end

denom(x::AbstractVector{T}, i::Int) where {T <: Number} =
  let j = firstindex(x) + i - 1
    mapreduce(subtracter(x[j]), *, SkipVec(x, j), init = one(T))
  end

# e must be zeroed. length(e) must be length(y) + 1.
#
# After the call, e[begin + i] corresponds to the i-th elementary
# symmetric polynomial.
#
# Based on the following recurrence:
#
#  e(k, n) = e(k - 1, n - 1) * x(n) + e(k, n - 1)
#
#  e(k, n) = 0  for n < k
#
#  e(0, n) = 1
function elementary_symmetric_polynomial_values!(
  e::AbstractVector{T},
  y::AbstractVector) where {T <: Number}

  e[begin] = one(T)

  local n = length(y)

  for i in 1:n
    for k in 1:(n - i + 1)
      local g = i + k - 1

      # TODO: MutableArithmetics
      e[begin + k] += e[begin + k - 1] * y[begin + g - 1]
    end
  end

  e
end

# The Vandermonde inverse
function vandermonde_inv!(inv::AbstractMatrix{T}, x::AbstractVector) where
{T <: Number}
  local n = length(x)
  local esps = zeros(T, n + 1)
  local den = zero(T)

  # Based on the formula for the inverse in Nick Higham's article on
  # the Vandermonde matrix from his What Is series:
  # https://github.com/higham/what-is

  for i in 1:n
    den = denom(x, i)

    fill!(esps, 0x0)
    elementary_symmetric_polynomial_values!(
      esps,
      SkipVec(x, firstindex(x) + i - 1))

    for j in 1:n
      local t = n - j
      inv[begin + j - 1, begin + i - 1] =
        esps[begin + t] * pow_minus_one(t) / den
    end
  end

  inv
end

vandermonde_inv(x::AbstractVector{T}) where {T <: Number} =
  vandermonde_inv!(
    let n = length(x)
      zeros(T, n, n)
    end,
    x)

end
