# Copyright © 2022 Neven Sajko

# Finds the same or equivalent constraints as would be produced by
# SymbolicGaussianEliminationVandermonde.vandermonde_important_parts,
# but does it by computing determinants of truncated augmented
# matrices instead of performing Gaussian elimination on the augmented
# matrix.
#
# This can be faster, because the time required for computing all the
# constraints/linear equations depends less strongly on the number of
# constraints/number of rows of the coefficient matrix. And the
# coefficient matrix is usually absurdly tall and thin.
#
# Furthermore, when no monomials are zeroed upfront (i.e. when the
# entire set of monomials, up to the highest degree monomial, is
# included), the coefficient matrix is a Vandermonde matrix (instead of
# just a generalized Vandermonde matrix). This is used to speed up
# computing the determinants, as there is a well known and simple
# formula for the determinant of a Vandermonde matrix.

module OverdeterminedSystemConsistencyConstraints

using LinearAlgebra, ..LinCombs, ..SkipVectors

export
  overdetermined_consistency_constraints,
  overdetermined_consistency_constraints_vandermonde,
  overdetermined_consistency_constraints_general

# Equivalent to `Int8(-1)^n` and `Int8(iseven(n) ? 1 : -1)`
pow_minus_one(n::Integer) =
  iseven(n)*Int8(2) - one(Int8)

function vandermonde_determinant(vec::AbstractVector{T}) where {T <: Number}
  local r = one(T)

  # TODO: MutableArithmetics
  for i in eachindex(vec)
    for j in (i + 1):lastindex(vec)
      r *= vec[j] - vec[i]
    end
  end

  r
end

function vandermonde_determinant_partial(vec::AbstractVector{T}, x::T) where
{T <: Number}

  local r = one(T)

  # TODO: MutableArithmetics
  for i in eachindex(vec)
    r *= x - vec[i]
  end

  r
end

function overdetermined_consistency_constraints_vandermonde(
  degree::S,
  poldom::AbstractVector{F}) where {S <: Signed, F <: Real}

  # In this case we should be able to do better by exploiting the fact
  # that our coefficient matrix is exactly a (tall, rectangular)
  # Vandermonde matrix, instead of just a generalized Vandermonde
  # matrix.
  #
  # Here the idea is that, for each row that would be a zero row after
  # Gaussian elimination, get the row's constraint/linear equation like
  # so: let n be the number of columns of the coefficient (Vandermonde)
  # matrix. Fix the first n rows of the augmented matrix, and turn it
  # into a square matrix by removing all rows below except for the one
  # row that we have chosen above (corresponding to the constraint).
  #
  # By the Rouché–Capelli theorem the determinant of the square matrix
  # must be zero (we want the linear system to have a single solution),
  # so an equivalent constraint/linear equation is obtained by equating
  # the determinant with zero.
  #
  # To obviate the need for symbolic computation, we first do a
  # Laplace expansion on the last column, where all the symbolic values
  # reside. In this way we get a new determinant (minor) for each
  # member of the last column. Each of these minors is a Vandermonde
  # determinant, determined by a simple well-known formula. All except
  # for the last minor will depend on the constraint, but we can mostly
  # evaluate each minor independently of the constraint, because of the
  # recursive way in which a Vandermonde determinant may be expressed,
  # relating it to the determinant of the same matrix but with any
  # single row and the last column removed.

  (degree < 0) && error("negative degree")

  local pol_coef_cnt = degree + one(S)

  local poldom_interp = @view poldom[begin:(begin - 1 + pol_coef_cnt)]

  local det_full = vandermonde_determinant(poldom_interp)

  local dets_without_last_row =
    F[vandermonde_determinant(SkipVec(poldom_interp, i))
      for i in eachindex(poldom_interp)]

  # TODO: MutableArithmetics
  LinComb{F}[[
    (==(i, pol_coef_cnt + 1) ?
      (Handle(row) => one(F)) :
      (Handle(i) => (
        dets_without_last_row[i] *
        vandermonde_determinant_partial(
          SkipVec(poldom_interp, firstindex(poldom_interp) - 1 + i),
          poldom[begin - 1 + row]) *
        pow_minus_one(pol_coef_cnt + i + 1) /
        det_full)))
    for i in 1:(pol_coef_cnt + 1)]
    for row in (pol_coef_cnt + 1):length(poldom)]
end

function overdetermined_consistency_constraints_general(
  monomials::AbstractVector{S},
  poldom::AbstractVector{F}) where {S <: Integer, F <: Real}

  # The idea is very similar to the one in the _vandermonde version,
  # except that now we do more Laplace expansion on the constraint row.
  #
  # And now we have to use `Base.det` instead of
  # `vandermonde_determinant`.

  local pol_coef_cnt = length(monomials)

  local poldom_interp = @view poldom[begin:(begin - 1 + pol_coef_cnt)]

  local ret =
    LinComb{F}[[
      (==(i, pol_coef_cnt + 1) ?
        (Handle(row) => one(F)) :
        (Handle(0) => zero(F)))
      for i in 1:(pol_coef_cnt + 1)]
      for row in (pol_coef_cnt + 1):length(poldom)]

  local coefficient_matrix_upper_square = F[x^j for x in poldom_interp, j in monomials]

  # Minor of minor of truncated augmented matrix determinant.
  local d = zero(F)

  # TODO: MutableArithmetics
  for i in 1:pol_coef_cnt
    local inds = SkipVec(1:pol_coef_cnt, i)

    for col in 1:pol_coef_cnt
      d =
        det(coefficient_matrix_upper_square[inds, SkipVec(1:pol_coef_cnt, col)]) *
        pow_minus_one(i + col + 1)

      for row in (pol_coef_cnt + 1):length(poldom)
        local (_, c) = ret[row - pol_coef_cnt][i]
        c += d * (poldom[begin - 1 + row])^(monomials[begin - 1 + col])
        ret[row - pol_coef_cnt][i] = Handle(i) => c
      end
    end
  end

  local det_full = det(coefficient_matrix_upper_square)

  # TODO: MutableArithmetics
  for i in 1:pol_coef_cnt
    for row in 1:(length(poldom) - pol_coef_cnt)
      local (_, c) = ret[row][i]
      c /= det_full
      ret[row][i] = Handle(i) => c
    end
  end

  ret
end

monomials_are_full(m::AbstractVector{T}) where {T <: Integer} =
  ==(m, map(T, 0:(length(m) - 1)))

overdetermined_consistency_constraints(
  monomials::AbstractVector{S},
  poldom::AbstractVector{F}) where {S <: Integer, F <: Real} =
  (monomials_are_full(monomials) ?
    # Coefficient matrix is a rectangular Vandermonde matrix.
    overdetermined_consistency_constraints_vandermonde(
      length(monomials) - 1,
      poldom) :
    # Coefficient matrix is merely a generalized Vandermonde matrix.
    overdetermined_consistency_constraints_general(
      monomials,
      poldom))

end
